// MouseRotation.cpp : Defines the entry point for the console application.
//

#include "stdio.h"
#include "windows.h"
#include "gl/gl.h"
#include "gl/glu.h"

// freeglut library
#define FREEGLUT_STATIC
#include "freeglut/glut.h"

#include "math.h"

// basic OpenGL program demonstrating how to handle mouse rotation of 
//  a simple mesh.  I am using the perspective view from 3ds Max as an
//  example of how the mouse should interact.

// when the user clicks and drags, we rotate the camera about our center
//  point.  

// controls
/*
	left-click + drag				- rotate about the center
	left-click + right-click + drag - move center
	mouse wheel						- zoom in and out
*/

/*	
	ROTATION
		- to rotate the cube, we simply call glRotatef with some parameter controlled by the mouse.
	TRANSLATION
		- to simulate a translation, we move the camera position AND the point it is looking at.
			This translation should be in the xy plane only(parallel to the monitor screen).
	ZOOM
		- to zoom, we have to shorten the distance between the camera position and the point the camera is 
			looking at.


*/

#ifndef M_PI
#define M_PI  3.14159265358979323846f
#endif

static float eyex, eyey, eyez;  // eye x,y,z values for gluLookAt (location of eye)
static float focusx, focusy, focusz; // the point the eye is looking at

struct g_mouseState{
	bool leftButton;
	bool rightButton;
	bool middleButton;
	int x;
	int y;
} MouseState;

#define DEGREES_PER_PIXEL 0.6f
#define RADIANS_PER_PIXEL 0.002f
#define UNITS_PER_PIXEL 0.01f
#define UNITS_PER_WHEELTICK 0.35f
#define ZOOM_FACTOR .04f
float g_xRotation, g_yRotation;

static float lmodel_ambient[4]    = {0.1f, 0.1f, 0.1f, 1.0f};


/*
 our display function, draws a box with proper rotation and 
 translation according to our mouse interaction.
*/
void Draw()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();

	// move our eye to the most recent place
	gluLookAt(eyex, eyey, eyez, focusx, focusy, focusz, 0,1,0);

	// rotate our box in the x and y directions
	glRotatef(g_xRotation, 0,1,0);
	glRotatef(g_yRotation, 1,0,0);

	// set our color to something
	glColor3f(.5f, .5f, .5f);
	// draw a cube
	glutSolidCube(1.0);

	glPopMatrix();
	glutSwapBuffers();
}

/*
 add a light to make the box we're showing a bit more interesting
*/
void InitializeLight()
{
	glMatrixMode(GL_MODELVIEW);
	
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);

	glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_CULL_FACE);
}

/*
  some basic initialization of our GL code
*/
void InitRenderer()
{
	glClearColor(0.0f, 0.0f, 1.0f, 0.0f);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// set up our viewing frustum
	gluPerspective(40.0f, 1.0f, 0.1f, 40.0f);

	// starting rotation.
	g_yRotation = -45.0f;
	g_xRotation = 30.0f;

	//initialize the mouse state
	MouseState.leftButton = MouseState.rightButton = MouseState.middleButton = false;
	MouseState.x = MouseState.y = 0;

	// init our eye location
	eyex = 0;
	eyey = 0;
	eyez = 5;
	focusx = focusy = focusz = 0.0f;

	// create some lighting
	InitializeLight();
}

void PrintMouseState()
{
	printf("mouse x: %d, y:%d left: %s, right:%s, middle:%s\n", MouseState.x,
		MouseState.y, MouseState.leftButton ? "down":"up", 
		MouseState.rightButton ? "down":"up", MouseState.middleButton ? "down":"up");
}

/*
  this function is called when any mouse buttons are pressed
*/
void HandleMouseState(int button, int state, int x, int y)
{
	// update our button state
	if(button == GLUT_LEFT_BUTTON)
	{
		if(state == GLUT_DOWN)
			MouseState.leftButton = true;
		else
			MouseState.leftButton = false;
	}
	if(button == GLUT_RIGHT_BUTTON)
	{
		if(state == GLUT_DOWN)
			MouseState.rightButton = true;
		else
			MouseState.rightButton = false;
	}
	if(button == GLUT_MIDDLE_BUTTON)
	{
		if(state == GLUT_DOWN)
			MouseState.middleButton = true;
		else
			MouseState.middleButton = false;
	}
	
	// wheel handlers (doesn't seem to be documented anywhere, but this works
	if(button == 3) // wheel up
	{
		// zooming.  we move our eye along the vector between the eye
		//  and the focus point.
		eyex = (1 - ZOOM_FACTOR) * eyex + focusx * ZOOM_FACTOR;
		eyey = (1 - ZOOM_FACTOR) * eyey + focusy * ZOOM_FACTOR;
		eyez = (1 - ZOOM_FACTOR) * eyez + focusz * ZOOM_FACTOR;
		glutPostRedisplay();
	}
	if(button == 4) // wheel down
	{
		eyex = (1 + ZOOM_FACTOR) * eyex - focusx * ZOOM_FACTOR;
		eyey = (1 + ZOOM_FACTOR) * eyey - focusy * ZOOM_FACTOR;
		eyez = (1 + ZOOM_FACTOR) * eyez - focusz * ZOOM_FACTOR;
		glutPostRedisplay();
	}

	// update our position so we know a delta when the mouse is moved
	MouseState.x = x;
	MouseState.y = y;

	PrintMouseState();
}

void HandleMouseMove(int x, int y)
{
	// calculate a delta in movement
	int yDelta = MouseState.y - y;
	int xDelta = MouseState.x - x;
	// commit the mouse position
	MouseState.x = x;
	MouseState.y = y;

	// when we need to rotate (only the left button is down)
	if(MouseState.leftButton && !MouseState.rightButton && !MouseState.middleButton)
	{
		// rotate by the delta
		g_xRotation -= xDelta * DEGREES_PER_PIXEL;
		g_yRotation -= yDelta * DEGREES_PER_PIXEL;
	}
	// if we need to move translate (left and right buttons are down
	else if(MouseState.leftButton && MouseState.rightButton && !MouseState.middleButton)
	{
		// move our eye
		eyex += xDelta * UNITS_PER_PIXEL;
		eyey -= yDelta * UNITS_PER_PIXEL;

		// move our focus point
		focusx += xDelta * UNITS_PER_PIXEL;
		focusy -= yDelta * UNITS_PER_PIXEL;
	}

	glutPostRedisplay();
	PrintMouseState();
}


int main(int argc, char* argv[])
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowSize(640,480);
	glutCreateWindow("Mouse Rotation");
	glutDisplayFunc(Draw);
	glutMouseFunc(HandleMouseState);
	glutMotionFunc(HandleMouseMove);

	InitRenderer();

	glutMainLoop();

	return 0;
}

